#!/usr/bin/env python

"""
Nonconformity functions.
"""

# Authors: Henrik Linusson

from __future__ import division

import numpy as np
from scipy.stats import pearsonr

# -----------------------------------------------------------------------------
# Classification error functions
# -----------------------------------------------------------------------------
def inverse_probability(prediction, y):
	"""Calculates the probability of not predicting the correct class.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		1 - \hat{P}(y_i | x) \, .

	Parameters
	----------
	prediction : numpy array of shape [n_samples, n_classes]
		Class probability estimates for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity scores of the samples.
	"""
	prob = np.zeros(y.size, dtype=np.float32)
	for i, y_ in enumerate(y):
		if y_ >= prediction.shape[1]:
			prob[i] = 0
		else:
			prob[i] = prediction[i, int(y_)]
	return 1 - prob

def margin(prediction, y):
	"""Calculates the margin error.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		0.5 - \hat{P}(y_i | x) - max_{y \, != \, y_i} \hat{P}(y | x) \, .

	Parameters
	----------
	prediction : numpy array of shape [n_samples, n_classes]
		Class probability estimates for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	Returns
	-------
	nc : numpy array of shape [n_samples, n_classes]
		Nonconformity scores for each sample and each class.
	"""
	prob = np.zeros(y.size, dtype=np.float32)
	for i, y_ in enumerate(y):
		if y_ >= prediction.shape[1]:
			prob[i] = 0
		else:
			prob[i] = prediction[i, int(y_)]
			prediction[i, int(y_)] = -np.inf
	return 0.5 - ((prob - prediction.max(axis=1)) / 2)

# -----------------------------------------------------------------------------
# Regression error functions
# -----------------------------------------------------------------------------
def abs_error(prediction, y, norm=None, beta=0):
	"""Calculates absolute error nonconformity for regression problems.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		| y_i - \hat{y}_i |

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Regression prediction for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	norm : numpy array of shape [n_samples]
		Normalization values for normalized nonconformity.

	beta : float
		Beta parameter for normalized nonconformity. Larger beta reduces
		influence of normalization model.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity score of each sample.
	"""
	norm = np.array([1] * prediction.shape[0]) if norm is None else norm
	beta = 0 if beta is None else beta
	return np.abs(prediction - y) / (norm + beta)

def abs_error_inv(prediction, nc, significance, norm=None, beta=0):
	"""Calculates a prediction from an absolute-error nonconformity function.

	Calculates the partial inverse of the ``absolute_error`` nonconformity
	function, i.e., the minimum and maximum boundaries of the prediction
	interval for each point prediction, given a significance level.

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Point (regression) predictions of a test examples.

	nc : numpy array of shape [n_calibration_samples]
		Nonconformity scores obtained for conformal predictor.

	significance : float
		Significance level (0, 1).

	norm : numpy array of shape [n_samples]
		Normalization values for normalized nonconformity.

	beta : float
		Beta parameter for normalized nonconformity. Larger beta reduces
		influence of normalization model.

	Returns
	-------
	interval : numpy array of shape [n_samples, 2]
		Minimum and maximum interval boundaries for each prediction.
	"""
	norm = np.array([1] * prediction.shape[0]) if norm is None else norm
	beta = 0 if beta is None else beta

	nc = np.sort(nc)[::-1]
	border = int(np.floor(significance * (nc.size + 1))) - 1
	# TODO: should probably warn against too few calibration examples
	border = min(max(border, 0), nc.size - 1)
	return np.vstack([prediction - (nc[border] * (norm + beta)),
	                  prediction + (nc[border] * (norm + beta))]).T

def sign_error(prediction, y, norm=None, beta=None):
	"""Calculates signed error nonconformity for regression problems.

	For each correct output in ``y``, nonconformity is defined as

	.. math::
		y_i - \hat{y}_i

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Regression prediction for each sample.

	y : numpy array of shape [n_samples]
		True output labels of each sample.

	norm : numpy array of shape [n_samples]
		Normalization values for normalized nonconformity.

	beta : float
		Beta parameter for normalized nonconformity. Larger beta reduces
		influence of normalization model.

	Returns
	-------
	nc : numpy array of shape [n_samples]
		Nonconformity score of each sample.
	"""
	norm = np.array([1] * prediction.shape[0]) if norm is None else norm
	beta = 0 if beta is None else beta

	return (prediction - y) / (norm + beta)

def sign_error_inv(prediction, nc, significance, norm=None, beta=None):
	"""Calculates a prediction from a signed-error nonconformity function.

	Calculates the partial inverse of the ``signed_error`` nonconformity
	function, i.e., the minimum and maximum boundaries of the prediction
	interval for each point prediction, given a significance level.

	Parameters
	----------
	prediction : numpy array of shape [n_samples]
		Point (regression) predictions of a test examples.

	nc : numpy array of shape [n_calibration_samples]
		Nonconformity scores obtained for conformal predictor.

	significance : float
		Significance level (0, 1).

	norm : numpy array of shape [n_samples]
		Normalization values for normalized nonconformity.

	beta : float
		Beta parameter for normalized nonconformity. Larger beta reduces
		influence of normalization model.

	Returns
	-------
	interval : numpy array of shape [n_samples, 2]
		Minimum and maximum interval boundaries for each prediction.
	"""
	norm = np.array([1] * prediction.shape[0]) if norm is None else norm
	beta = 0 if beta is None else beta

	nc = np.sort(nc)[::-1]
	upper = int(np.floor((significance / 2) * (nc.size + 1)))
	lower = int(np.floor((1 - significance / 2) * (nc.size + 1)))
	# TODO: should probably warn against too few calibration examples
	upper = min(max(upper, 0), nc.size - 1)
	lower = max(min(lower, nc.size - 1), 0)
	return np.vstack([prediction + (nc[lower] * (norm + beta)),
	                  prediction + (nc[upper] * (norm + beta))]).T

# -----------------------------------------------------------------------------
# Base nonconformity scorer
# -----------------------------------------------------------------------------
class BaseNc(object):
	"""Base class for nonconformity scorers based on an underlying model.
	"""
	def __init__(self, model_class, err_func, model_params=None):
		self.err_func = err_func
		self.model_class = model_class
		self.model_params = model_params if model_params else {}
		self.model = self.model_class(**self.model_params)

		self.last_x, self.last_y = None, None
		self.last_prediction = None
		self.clean = False

	def fit(self, x, y):
		"""Fits the underlying model of the nonconformity scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the underlying model.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the underlying model.

		Returns
		-------
		None
		"""
		self.model.fit(x, y)
		self.clean = False

	def __get_prediction(self, x):
		if (not self.clean or
			self.last_x is None or
		    not np.array_equal(self.last_x, x)):

			self.last_x = x
			self.last_prediction = self._underlying_predict(x)
			self.clean = True

		return self.last_prediction.copy()

	def calc_nc(self, x, y):
		"""Calculates the nonconformity score of a set of samples.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for which to calculate a nonconformity score.

		y : numpy array of shape [n_samples, n_features]
			Outputs of examples for which to calculate a nonconformity score.

		Returns
		-------
		nc : numpy array of shape [n_samples]
			Nonconformity scores of samples.
		"""
		prediction = self.__get_prediction(x)
		return self.err_func(prediction, y)

# -----------------------------------------------------------------------------
# Classification nonconformity scorers
# -----------------------------------------------------------------------------
class ProbEstClassifierNc(BaseNc):
	"""Nonconformity scorer using an underlying class probability estimating
	model.

	Parameters
	----------
	model_class : class
		The model_class should be implement the ``fit(x, y)`` and
		``predict_proba(x)`` methods, as used by the classification models
		present in the scikit-learn library.

	err_func : callable
		Scorer callable object with signature ``score(estimator, x, y)``.

	model_params : dict, optional
		Dict containing keyword parameters to pass to model_class upon
		initialization.

	Attributes
	----------
	model_class : class
		Class used to construct the underlying model.

	err_func : callable
		Scorer function used to calculate nonconformity scores.

	model_params : dict
		Parameters sent to underlying model.

	model : object
		Underlying model object.

	See also
	--------
	RegressorNc
	"""
	def __init__(self,
	             model_class,
	             err_func=margin,
	             model_params=None):
		super(ProbEstClassifierNc, self).__init__(model_class,
		                                          err_func,
		                                          model_params)

	def _underlying_predict(self, x):
		return self.model.predict_proba(x)

# -----------------------------------------------------------------------------
# Regression nonconformity scorers
# -----------------------------------------------------------------------------
class RegressorNc(BaseNc):
	"""Nonconformity scorer using an underlying regression model.

	Parameters
	----------
	model_class : class
		The model_class should be implement the ``fit(x, y)`` and
		``predict(x)`` methods, as used by the regression models
		present in the scikit-learn library.

	err_func : callable
		Scorer callable object with signature ``score(estimator, x, y)``.

	inverse_error_func : callable
		Inverse (or partial inverse) of err_func.

	model_params : dict, optional
		Dict containing keyword parameters to pass to model_class upon
		initialization.

	Attributes
	----------
	model_class : class
		Class used to construct the underlying model.

	err_func : callable
		Scorer function used to calculate nonconformity scores.

	inverse_err_func : callable
		Inverse function (partial) of nonconformity function.

	model_params : dict
		Parameters sent to underlying model.

	model : object
		Underlying model object.

	See also
	--------
	ProbEstClassifierNc
	"""
	def __init__(self,
	             model_class,
	             err_func=abs_error,
	             inverse_err_func=abs_error_inv,
	             model_params=None):
		super(RegressorNc, self).__init__(model_class,
		                                  err_func,
		                                  model_params)

		self.inverse_err_func = inverse_err_func

	def _underlying_predict(self, x):
		return self.model.predict(x)

	def predict(self, x, nc, significance=None):
		"""Constructs prediction intervals for a set of test examples.

		Predicts the output of each test pattern using the underlying model,
		and applies the (partial) inverse nonconformity function to each
		prediction, resulting in a prediction interval for each test pattern.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then intervals for
			all significance levels (0.01, 0.02, ..., 0.99) are output in a
			3d-matrix.

		Returns
		-------
		p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99}
			If significance is ``None``, then p contains the interval (minimum
			and maximum boundaries) for each test pattern, and each significance
			level (0.01, 0.02, ..., 0.99). If significance is a float between
			0 and 1, then p contains the prediction intervals (minimum and
			maximum	boundaries) for the set of test patterns at the chosen
			significance level.
		"""
		prediction = self._underlying_predict(x)
		if significance:
			return self.inverse_err_func(prediction, nc, significance)
		else:
			significance = np.arange(0.01, 1.0, 0.01)
			return np.dstack([self.inverse_err_func(prediction, nc, s)
			                  for s in significance])

class NormalizedRegressorNc(RegressorNc):
	def __init__(self,
	             model_class,
	             normalizer_class,
	             err_func=abs_error,
	             inverse_err_func=abs_error_inv,
	             model_params=None,
	             normalizer_params=None,
	             beta='auto'):
		super(NormalizedRegressorNc, self).__init__(model_class,
		                                            err_func,
		                                            inverse_err_func,
		                                            model_params)
		self.normalizer_params = normalizer_params if normalizer_params else {}
		self.normalizer_class = normalizer_class
		self.normalizer = self.normalizer_class(**self.normalizer_params)
		self.beta = beta
		self.beta_ = None

	def fit(self, x, y):
		super(NormalizedRegressorNc, self).fit(x, y)
		err = np.abs(self._underlying_predict(x) - y)
		err += 0.00001 # Add a small error to each sample to avoid log(0)
		log_err = np.log(err)
		self.normalizer.fit(x, log_err)

	def calc_nc(self, x, y):
		norm = np.exp(self.normalizer.predict(x))
		prediction = self._underlying_predict(x)

		if self.beta == 'auto':
			r = pearsonr(norm, prediction)[0]
			if r < 0:
				self.beta_ = 0
			elif r < 0.3:
				self.beta_ = 1
			else:
				self.beta_ = 10
		else:
			self.beta_ = self.beta

		return self.err_func(prediction, y, norm, self.beta_)

	def predict(self, x, nc, significance=None):
		prediction = self._underlying_predict(x)
		norm = np.exp(self.normalizer.predict(x))
		if significance:
			return self.inverse_err_func(prediction,
			                             nc,
			                             significance,
			                             norm,
			                             self.beta_)
		else:
			significance = np.arange(0.01, 1.0, 0.01)
			return np.dstack([self.inverse_err_func(prediction,
			                                        nc,
			                                        s,
			                                        norm,
			                                        self.beta_)
			                  for s in significance])



#!/usr/bin/env python

"""
Inductive conformal predictors.
"""

# Authors: Henrik Linusson

from __future__ import division
from collections import defaultdict
from functools import partial

import numpy as np

from nonconformist.base import RegressorMixin, ClassifierMixin

# -----------------------------------------------------------------------------
# Base inductive conformal predictor
# -----------------------------------------------------------------------------
class BaseIcp(object):
	"""Base class for inductive conformal predictors.
	"""
	def __init__(self, nc_function, condition=None):
		self.cal_x, self.cal_y = None, None
		self.nc_function = nc_function
		self.condition = condition if condition else lambda x: 0

	def fit(self, x, y):
		"""Fit underlying nonconformity scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for fitting the nonconformity scorer.

		y : numpy array of shape [n_samples]
			Outputs of examples for fitting the nonconformity scorer.

		Returns
		-------
		None
		"""
		#TODO: incremental?
		self.nc_function.fit(x, y)

	def calibrate(self, x, y, increment=False):
		"""Calibrate conformal predictor based on underlying nonconformity
		scorer.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of examples for calibrating the conformal predictor.

		y : numpy array of shape [n_samples, n_features]
			Outputs of examples for calibrating the conformal predictor.

		increment : boolean
			If ``True``, performs an incremental recalibration of the conformal
			predictor. The supplied ``x`` and ``y`` are added to the set of
			previously existing calibration examples, and the conformal
			predictor is then calibrated on both the old and new calibration
			examples.

		Returns
		-------
		None
		"""
		# TODO: conditional
		self._calibrate_hook(x, y, increment)
		self._update_calibration_set(x, y, increment)
		condition_map = np.array([self.condition((x[i, :], y[i]))
		                          for i in range(y.size)])
		self.conditions = np.unique(condition_map)
		self.cal_scores = defaultdict(partial(np.ndarray, 0))
		for cond in self.conditions:
			idx = condition_map == cond
			self.cal_scores[cond] = self.nc_function.calc_nc(self.cal_x[idx, :],
			                                                 self.cal_y[idx])

	def _calibrate_hook(self, x, y, increment):
		pass

	def _update_calibration_set(self, x, y, increment):
		if increment and self.cal_x is not None and self.cal_y is not None:
			self.cal_x = np.vstack([self.cal_x, x])
			self.cal_y = np.hstack([self.cal_y, y])
		else:
			self.cal_x, self.cal_y = x, y

	def get_params(self, deep=False):
		return {'nc_function': self.nc_function,
		        'condition': self.condition}

# -----------------------------------------------------------------------------
# Inductive conformal classifier
# -----------------------------------------------------------------------------
class IcpClassifier(BaseIcp, ClassifierMixin):
	"""Inductive conformal classifier.

	Parameters
	----------
	nc_function : object
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``
		and ``calc_nc(x, y)``.

	smoothing : boolean
		Decides whether to use stochastic smoothing of p-values.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : object
		Nonconformity scorer object used to calculate nonconformity scores.

	See also
	--------
	IcpRegressor

	References
	----------
	.. [1] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.tree import DecisionTreeClassifier
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ProbEstClassifierNc, margin
	>>> iris = load_iris()
	>>> idx = np.random.permutation(iris.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> nc = ProbEstClassifierNc(DecisionTreeClassifier, margin)
	>>> icp = IcpClassifier(nc)
	>>> icp.fit(iris.data[train, :], iris.target[train])
	>>> icp.calibrate(iris.data[cal, :], iris.target[cal])
	>>> icp.predict(iris.data[test, :], significance=0.10)
	...             # doctest: +SKIP
	array([[ True, False, False],
		[False,  True, False],
		...,
		[False,  True, False],
		[False,  True, False]], dtype=bool)
	"""
	def __init__(self, nc_function, condition=None, smoothing=True):
		super(IcpClassifier, self).__init__(nc_function, condition)
		self.classes = None
		self.last_p = None
		self.smoothing = smoothing

	def _calibrate_hook(self, x, y, increment=False):
		self._update_classes(y, increment)

	def _update_classes(self, y, increment):
		if self.classes is None or not increment:
			self.classes = np.unique(y)
		else:
			self.classes = np.unique(np.hstack([self.classes, y]))

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float or None
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then the p-values
			are output rather than the predictions.

		Returns
		-------
		p : numpy array of shape [n_samples, n_classes]
			If significance is ``None``, then p contains the p-values for each
			sample-class pair; if significance is a float between 0 and 1, then
			p is a boolean array denoting which labels are included in the
			prediction sets.
		"""
		# TODO: if x == self.last_x ...
		n_test_objects = x.shape[0]
		p = np.zeros((n_test_objects, self.classes.size))
		for i, c in enumerate(self.classes):
			test_class = np.zeros(x.shape[0])
			test_class.fill(c)

			# TODO: maybe calculate p-values using cython or similar
			# TODO: interpolated p-values

			test_nc_scores = self.nc_function.calc_nc(x, test_class)
			for j, nc in enumerate(test_nc_scores):
				cal_scores = self.cal_scores[self.condition((x[j, :], c))]
				n_cal = cal_scores.size
				n_gt = np.sum(cal_scores > nc)
				n_eq = np.sum(cal_scores == nc) + 1
				p[j, i] = n_gt / (n_cal + 1)

				if self.smoothing:
					p[j, i] += (n_eq * np.random.uniform(0, 1, 1)) / (n_cal + 1)
				else:
					p[j, i] += n_eq / (n_cal + 1)

		if significance:
			return p > significance
		else:
			return p

# -----------------------------------------------------------------------------
# Inductive conformal regressor
# -----------------------------------------------------------------------------
class IcpRegressor(BaseIcp, RegressorMixin):
	"""Inductive conformal regressor.

	Parameters
	----------
	nc_function : object
		Nonconformity scorer object used to calculate nonconformity of
		calibration examples and test patterns. Should implement ``fit(x, y)``,
		``calc_nc(x, y)`` and ``predict(x, nc_scores, significance)``.

	Attributes
	----------
	cal_x : numpy array of shape [n_cal_examples, n_features]
		Inputs of calibration set.

	cal_y : numpy array of shape [n_cal_examples]
		Outputs of calibration set.

	nc_function : object
		Nonconformity scorer object used to calculate nonconformity scores.

	See also
	--------
	IcpClassifier

	References
	----------
	.. [1] Papadopoulos, H., Proedrou, K., Vovk, V., & Gammerman, A. (2002).
		Inductive confidence machines for regression. In Machine Learning: ECML
		2002 (pp. 345-356). Springer Berlin Heidelberg.

	.. [2] Papadopoulos, H., & Haralambous, H. (2011). Reliable prediction
		intervals with regression neural networks. Neural Networks, 24(8),
		842-851.

	Examples
	--------
	>>> import numpy as np
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.tree import DecisionTreeRegressor
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, abs_error, abs_error_inv
	>>> boston = load_boston()
	>>> idx = np.random.permutation(boston.target.size)
	>>> train = idx[:int(idx.size / 3)]
	>>> cal = idx[int(idx.size / 3):int(2 * idx.size / 3)]
	>>> test = idx[int(2 * idx.size / 3):]
	>>> nc = RegressorNc(DecisionTreeRegressor, abs_error, abs_error_inv)
	>>> icp = IcpRegressor(nc)
	>>> icp.fit(boston.data[train, :], boston.target[train])
	>>> icp.calibrate(boston.data[cal, :], boston.target[cal])
	>>> icp.predict(boston.data[test, :], significance=0.10)
	...     # doctest: +SKIP
	array([[  5. ,  20.6],
		[ 15.5,  31.1],
		...,
		[ 14.2,  29.8],
		[ 11.6,  27.2]])
	"""
	def __init__(self, nc_function, condition=None):
		super(IcpRegressor, self).__init__(nc_function, condition)

	def predict(self, x, significance=None):
		"""Predict the output values for a set of input patterns.

		Parameters
		----------
		x : numpy array of shape [n_samples, n_features]
			Inputs of patters for which to predict output values.

		significance : float
			Significance level (maximum allowed error rate) of predictions.
			Should be a float between 0 and 1. If ``None``, then intervals for
			all significance levels (0.01, 0.02, ..., 0.99) are output in a
			3d-matrix.

		Returns
		-------
		p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99}
			If significance is ``None``, then p contains the interval (minimum
			and maximum boundaries) for each test pattern, and each significance
			level (0.01, 0.02, ..., 0.99). If significance is a float between
			0 and 1, then p contains the prediction intervals (minimum and
			maximum	boundaries) for the set of test patterns at the chosen
			significance level.
		"""
		# TODO: interpolated p-values

		n_significance = (99 if significance is None
		                  else np.array(significance).size)

		if n_significance > 1:
			prediction = np.zeros((x.shape[0], 2, n_significance))
		else:
			prediction = np.zeros((x.shape[0], 2))

		condition_map = np.array([self.condition((x[i, :], None))
		                          for i in range(x.shape[0])])

		for condition in self.conditions:
			idx = condition_map == condition
			if np.sum(idx) > 0:
				p = self.nc_function.predict(x[idx, :],
				                             self.cal_scores[condition],
				                             significance)
				if n_significance > 1:
					prediction[idx, :, :] = p
				else:
					prediction[idx, :] = p

		return prediction


#!/usr/bin/env python

from distutils.core import setup
import nonconformist

setup(
	name = 'nonconformist',
	packages = ['nonconformist'],
	version = nonconformist.__version__,
	description = 'Python implementation of the conformal prediction framework.',
	author = 'Henrik Linusson',
	author_email = 'henrik.linusson@gmail.com',
	url = 'https://github.com/donlnz/nonconformist',
	download_url = 'https://github.com/donlnz/nonconformist/tarball/' + nonconformist.__version__,
	install_requires = ['numpy', 'scikit-learn'],
	keywords = ['conformal prediction',
	            'machine learning',
	            'classification',
	            'regression'],
	classifiers=['Intended Audience :: Science/Research',
	             'Intended Audience :: Developers',
                 'Programming Language :: Python',
                 'Topic :: Software Development',
                 'Topic :: Scientific/Engineering',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: Unix',
                 'Operating System :: MacOS',
                 ],
)

# Authors: Henrik Linusson



#!/usr/bin/env python

"""
Evaluation of conformal predictors.
"""

# Authors: Henrik Linusson

# TODO: cross_val_score/run_experiment should possibly allow multiple to be evaluated on identical folding

from __future__ import division

import numpy as np

from sklearn.cross_validation import StratifiedShuffleSplit
from sklearn.cross_validation import KFold
from sklearn.cross_validation import train_test_split

from nonconformist.base import RegressorMixin, ClassifierMixin

class BaseIcpCvHelper(object):
	"""Base class for cross validation helpers.
	"""
	def __init__(self, icp, calibration_portion):
		self.icp = icp
		self.cal_port = calibration_portion

	def predict(self, x, significance=None):
			return self.icp.predict(x, significance)

	def get_params(self, deep=False):
		return {'icp': self.icp,
		        'calibration_portion': self.cal_port}

class ClassIcpCvHelper(BaseIcpCvHelper, ClassifierMixin):
	"""Helper class for running the ``cross_val_score`` evaluation
	method on IcpClassifiers.

	See also
	--------
	IcpRegCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_iris
	>>> from sklearn.ensemble import RandomForestClassifier
	>>> from nonconformist.icp import IcpClassifier
	>>> from nonconformist.nc import ProbEstClassifierNc, margin
	>>> from nonconformist.evaluation import ClassIcpCvHelper
	>>> from nonconformist.evaluation import class_mean_errors
	>>> from nonconformist.evaluation import cross_val_score
	>>> data = load_iris()
	>>> nc = ProbEstClassifierNc(RandomForestClassifier, margin)
	>>> icp = IcpClassifier(nc)
	>>> icp_cv = ClassIcpCvHelper(icp)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 iterations=2,
	...                 folds=2,
	...                 scoring_funcs=[class_mean_errors],
	...                 significance_levels=[0.1])
	...     # doctest: +SKIP
	   class_mean_errors  fold  iter  significance
	0           0.013333     0     0           0.1
	1           0.080000     1     0           0.1
	2           0.053333     0     1           0.1
	3           0.080000     1     1           0.1
	"""
	def __init__(self, icp, calibration_portion=0.25):
		super(ClassIcpCvHelper, self).__init__(icp, calibration_portion)

	def fit(self, x, y):
		split = StratifiedShuffleSplit(y, n_iter=1, test_size=self.cal_port)
		for train, cal in split:
			self.icp.fit(x[train, :], y[train])
			self.icp.calibrate(x[cal, :], y[cal])

class RegIcpCvHelper(BaseIcpCvHelper, RegressorMixin):
	"""Helper class for running the ``cross_val_score`` evaluation
	method on IcpRegressors.

	See also
	--------
	IcpClassCrossValHelper

	Examples
	--------
	>>> from sklearn.datasets import load_boston
	>>> from sklearn.ensemble import RandomForestRegressor
	>>> from nonconformist.icp import IcpRegressor
	>>> from nonconformist.nc import RegressorNc, abs_error, abs_error_inv
	>>> from nonconformist.evaluation import RegIcpCvHelper
	>>> from nonconformist.evaluation import reg_mean_errors
	>>> from nonconformist.evaluation import cross_val_score
	>>> data = load_boston()
	>>> nc = RegressorNc(RandomForestRegressor, abs_error, abs_error_inv)
	>>> icp = IcpRegressor(nc)
	>>> icp_cv = RegIcpCvHelper(icp)
	>>> cross_val_score(icp_cv,
	...                 data.data,
	...                 data.target,
	...                 iterations=2,
	...                 folds=2,
	...                 scoring_funcs=[reg_mean_errors],
	...                 significance_levels=[0.1])
	...     # doctest: +SKIP
	   fold  iter  reg_mean_errors  significance
	0     0     0         0.185771           0.1
	1     1     0         0.138340           0.1
	2     0     1         0.071146           0.1
	3     1     1         0.043478           0.1
	"""
	def __init__(self, icp, calibration_portion=0.25):
		super(RegIcpCvHelper, self).__init__(icp, calibration_portion)

	def fit(self, x, y):
		split = train_test_split(x, y, test_size=self.cal_port)
		x_tr, x_cal, y_tr, y_cal = split[0], split[1], split[2], split[3]
		self.icp.fit(x_tr, y_tr)
		self.icp.calibrate(x_cal, y_cal)

# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
from sklearn.base import clone
import pandas
import sys

def cross_val_score(model, x, y, iterations=10, folds=10, fit_params=None,
					scoring_funcs=None, significance_levels=None,
					verbose=False):
	"""Evaluates a conformal predictor using cross-validation.

	Parameters
	----------
	model : object
		Conformal predictor to evaluate.

	x : numpy array of shape [n_samples, n_features]
		Inputs of data to use for evaluation.

	y : numpy array of shape [n_samples]
		Outputs of data to use for evaluation.

	iterations : int
		Number of iterations to use for evaluation. The data set is randomly
		shuffled before each iteration.

	folds : int
		Number of folds to use for evaluation.

	fit_params : dictionary
		Parameters to supply to the conformal prediction object on training.

	scoring_funcs : iterable
		List of evaluation functions to apply to the conformal predictor in each
		fold. Each evaluation function should have a signature
		``scorer(prediction, y, significance)``.

	significance_levels : iterable
		List of significance levels at which to evaluate the conformal
		predictor.

	verbose : boolean
		Indicates whether to output progress information during evaluation.

	Returns
	-------
	scores : pandas DataFrame
		Tabulated results for each iteration, fold and evaluation function.
	"""

	fit_params = fit_params if fit_params else {}
	significance_levels = (significance_levels if significance_levels
	                       is not None else np.arange(0.01, 1.0, 0.01))

	df = pandas.DataFrame()

	columns = ['iter',
			   'fold',
			   'significance',
			   ] + [f.__name__ for f in scoring_funcs]
	for i in range(iterations):
		idx = np.random.permutation(y.size)
		x, y = x[idx, :], y[idx]
		cv = KFold(y.size, folds)
		for j, (train, test) in enumerate(cv):
			if verbose:
				sys.stdout.write('\riter {}/{} fold {}/{}'.format(
					i + 1,
					iterations,
					j + 1,
					folds
				))
			m = clone(model)
			m.fit(x[train, :], y[train], **fit_params)
			prediction = m.predict(x[test, :], significance=None)
			for k, s in enumerate(significance_levels):
				scores = [scoring_func(prediction, y[test], s)
						  for scoring_func in scoring_funcs]
				df_score = pandas.DataFrame([[i, j, s] + scores],
											columns=columns)
				df = df.append(df_score, ignore_index=True)

	return df

def run_experiment(model, csv_files, iterations=10, folds=10, fit_params=None,
				   scoring_funcs=None, significance_levels=None,
				   normalize=False, verbose=False):
	"""Performs a cross-validation evaluation of a conformal predictor on a
	collection of data sets in csv format.

	Parameters
	----------
	model : object
		Conformal predictor to evaluate.

	csv_files : iterable
		List of file names (with absolute paths) containing csv-data, used to
		evaluate the conformal predictor.

	iterations : int
		Number of iterations to use for evaluation. The data set is randomly
		shuffled before each iteration.

	folds : int
		Number of folds to use for evaluation.

	fit_params : dictionary
		Parameters to supply to the conformal prediction object on training.

	scoring_funcs : iterable
		List of evaluation functions to apply to the conformal predictor in each
		fold. Each evaluation function should have a signature
		``scorer(prediction, y, significance)``.

	significance_levels : iterable
		List of significance levels at which to evaluate the conformal
		predictor.

	verbose : boolean
		Indicates whether to output progress information during evaluation.

	Returns
	-------
	scores : pandas DataFrame
		Tabulated results for each data set, iteration, fold and
		evaluation function.
	"""
	df = pandas.DataFrame()
	is_regression = model.get_problem_type() == 'regression'

	n_data_sets = len(csv_files)
	for i, csv_file in enumerate(csv_files):
		if verbose:
			print('\n{} ({} / {})'.format(csv_file, i + 1, n_data_sets))
		data = pandas.read_csv(csv_file)
		x, y = data.values[:, :-1], data.values[:, -1]
		if normalize:
			if is_regression:
				y = y - y.min() / (y.max() - y.min())
			else:
				for i, y_ in enumerate(np.unique(y)):
					y[y == y_] = i

		scores = cross_val_score(model, x, y, iterations, folds, fit_params,
								 scoring_funcs, significance_levels, verbose)

		ds_df = pandas.DataFrame(scores)
		try:
			ds_df['data_set'] = csv_file.split('/')[-1]
		except:
			ds_df['data_set'] = csv_file

		df = df.append(ds_df)

	return df

# -----------------------------------------------------------------------------
# Validity measures
# -----------------------------------------------------------------------------
def reg_n_correct(prediction, y, significance):
	"""Calculates the number of correct predictions made by a conformal
	regression model.
	"""
	idx = int(significance * 100 - 1)
	prediction = prediction[:, :, idx]
	low = y >= prediction[:,0]
	high = y <= prediction[:,1]
	correct = low * high

	return y[correct].size

def reg_mean_errors(prediction, y, significance):
	"""Calculates the average error rate of a conformal regression model.
	"""
	return 1 - reg_n_correct(prediction, y, significance) / y.size

def class_n_correct(prediction, y, significance):
	"""Calculates the number of correct predictions made by a conformal
	classification model.
	"""
	prediction = prediction > significance
	correct = np.zeros((y.size,), dtype=bool)
	for i, y_ in enumerate(y):
		correct[i] = prediction[i, y_]
	return np.sum(correct)

def class_mean_errors(prediction, y, significance=None):
	"""Calculates the average error rate of a conformal classification model.
	"""
	return 1 - (class_n_correct(prediction, y, significance) / y.size)

# -----------------------------------------------------------------------------
# Efficiency measures
# -----------------------------------------------------------------------------
def _reg_interval_size(prediction, y, significance):
	idx = int(significance * 100 - 1)
	prediction = prediction[:, :, idx]

	return prediction[:, 1] - prediction[:, 0]

def reg_min_size(prediction, y, significance):
	return np.min(_reg_interval_size(prediction, y, significance))

def reg_q1_size(prediction, y, significance):
	return np.percentile(_reg_interval_size(prediction, y, significance), 25)

def reg_median_size(prediction, y, significance):
	return np.median(_reg_interval_size(prediction, y, significance))

def reg_q3_size(prediction, y, significance):
	return np.percentile(_reg_interval_size(prediction, y, significance), 75)

def reg_max_size(prediction, y, significance):
	return np.max(_reg_interval_size(prediction, y, significance))

def reg_mean_size(prediction, y, significance):
	"""Calculates the average prediction interval size of a conformal
	regression model.
	"""
	return np.mean(_reg_interval_size(prediction, y, significance))

def class_avg_c(prediction, y, significance):
	"""Calculates the average number of classes per prediction of a conformal
	classification model.
	"""
	prediction = prediction > significance
	return np.sum(prediction) / prediction.shape[0]

def class_mean_p_val(prediction, y, significance):
	"""Calculates the mean of the p-values output by a conformal classification
	model.
	"""
	return np.mean(prediction)

def class_one_c(prediction, y, significance):
	"""Calculates the rate of singleton predictions (prediction sets containing
	only a single class label) of a conformal classification model.
	"""
	prediction = prediction > significance
	return np.sum(1 for _ in filter(lambda x: np.sum(x) == 1, prediction))


#!/usr/bin/env python

"""
docstring
"""

# Authors: Henrik Linusson

class RegressorMixin(object):
	@classmethod
	def get_problem_type(self):
		return 'regression'

class ClassifierMixin(object):
	@classmethod
	def get_problem_type(self):
		return 'classification'